/*
	Copyright (C) 2010 Hans-Werner Hilse <hilse@web.de>
	See file "COPYING" for licence information.
*/
#include <sys/socket.h>
#include <sys/un.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "vipipc.h"
#include "vipipc_prototypes.h"

#ifdef DEBUG
#include <stdio.h>
void hex_dump(void *data, int size) {
    /* dumps size bytes of *data to stdout. Looks like:
     * [0000]  75 6E 6B 6E 6F 77 6E 20  30 FF 00 00 00 00 39 00  unknown0.....9.
     *
     * stolen from http://sws.dett.de/mini/hexdump-c/hexdump.c
     * (public domain licence), a bit adapted
     */

    unsigned char *p = data;
    unsigned char c;
    int n;
    char bytestr[4] = {0};
    char addrstr[10] = {0};
    char hexstr[ 16*3 + 5] = {0};
    char charstr[16*1 + 5] = {0};
    for(n=1;n<=size;n++) {
        if (n%16 == 1) {
            /* store address for this line */
            snprintf(addrstr, sizeof(addrstr), "%.4x",
               ((unsigned int)p-(unsigned int)data) );
        }
            
        c = *p;
        if (isprint(c) == 0) {
            c = '.';
        }

        /* store hex str (for left side) */
        snprintf(bytestr, sizeof(bytestr), "%02X ", *p);
        strncat(hexstr, bytestr, sizeof(hexstr)-strlen(hexstr)-1);

        /* store char str (for right side) */
        snprintf(bytestr, sizeof(bytestr), "%c", c);
        strncat(charstr, bytestr, sizeof(charstr)-strlen(charstr)-1);

        if(n%16 == 0) { 
            /* line completed */
            printf("IPC: [%4.4s]  %-48.48s  %s\n", addrstr, hexstr, charstr);
            hexstr[0] = 0;
            charstr[0] = 0;
        } else if(n%8 == 0) {
            /* half line: add whitespaces */
            strncat(hexstr, " ", sizeof(hexstr)-strlen(hexstr)-1);
        }
        p++; /* next byte */
    }

    if (strlen(hexstr) > 0) {
        /* print rest of buffer if not empty */
        printf("IPC: [%4.4s]  %-48.48s  %s\n", addrstr, hexstr, charstr);
    }
}
#endif /* DEBUG */

/* create a new connection "object" */

extern ipcconn ipcconn_new(const char* socketname, const char* interfacename) {
	ipcconn conn = (ipcconn) malloc(sizeof(ipcconn_s));
	if(conn) {
		conn->socketname = strdup(socketname);
		if(conn->socketname) {
			conn->interfacename = strdup(interfacename);
			if(conn->interfacename) {
				conn->fd = socket(PF_UNIX, SOCK_STREAM, 0);
				if(conn->fd >= 0) {
					conn->lazy_connect = 1;
					return conn; /* successfully initialized */
				}
				free(conn->interfacename);
			}
			free(conn->socketname);
		}
		free(conn);
	}
	return NULL;
}

/* free resources & close open socket for connection */

extern void ipcconn_free(ipcconn* conn) {
	if(!*conn)
		return;
	if((*conn)->fd >= 0)
		ipcconn_close(*conn);
	if((*conn)->interfacename)
		free((*conn)->interfacename);
	if((*conn)->socketname)
		free((*conn)->socketname);
	free(*conn);
	*conn = NULL;
}

/* simple helper to push connection info and an RPC call onto a ipcmessage stack */

extern int ipcconn_message_push_rpc(ipcconn conn, ipcmessage msg, const char* procedure, const char statusbyte) {
	if(!conn || !conn->interfacename || !conn->socketname)
		return 1;
	if(ipcmessage_push_zstring(msg, conn->socketname)
		|| ipcmessage_push_zstring(msg, conn->interfacename)
		|| ipcmessage_push_zstring(msg, procedure)
		|| ipcmessage_push_byte(msg, statusbyte))
		return 1;
	return 0;
}

/* socket management: */

extern int ipcconn_connect(ipcconn client) {
	struct sockaddr_un address;
	size_t address_len;
	char* hash;

	if(!client || !client->socketname || client->fd < 0)
		return 1;

	address.sun_family = AF_UNIX;
	strncpy(address.sun_path, client->socketname, sizeof(address.sun_path));
	address.sun_path[sizeof(address.sun_path)-1] = '\0';
	hash = strrchr(address.sun_path, '#');
	if(hash > strrchr(address.sun_path, '/'))
		*hash = '\0';
	address_len = sizeof(address.sun_family) + strlen(address.sun_path);
	return connect(client->fd, (struct sockaddr*) &address, address_len);
}

extern int ipcconn_bind(ipcconn server, mode_t socketpermissions) {
	struct sockaddr_un address;
	size_t address_len;
	char* hash;
	int err;

	if(!server || !server->socketname || server->fd < 0)
		return 1;

	server->lazy_connect = 0;

	address.sun_family = AF_UNIX;
	strncpy(address.sun_path, server->socketname, sizeof(address.sun_path));
	address.sun_path[sizeof(address.sun_path)-1] = '\0';
	hash = strrchr(address.sun_path, '#');
	if(hash > strrchr(address.sun_path, '/'))
		*hash = '\0';
	address_len = sizeof(address.sun_family) + strlen(address.sun_path);

	unlink(address.sun_path);

	err = bind(server->fd, (struct sockaddr*) &address, address_len);
	if(!err) chmod(address.sun_path, socketpermissions);
	return err;
}

extern int ipcconn_listen(ipcconn server, int backlog) {
	if(!server || server->fd < 0)
		return 1;
	return listen(server->fd, backlog);
}

extern int ipcconn_bind_and_listen(ipcconn server, mode_t socketpermissions, int backlog) {
	int err = ipcconn_bind(server, socketpermissions);
	if(!err)
		err = ipcconn_listen(server, backlog);
	return err;
}

extern int ipcconn_accept(ipcconn server, ipcconn *client) {
	int fd_client;

	if(!server || !client)
		return 1;

	fd_client = accept(server->fd, NULL, NULL);
	if(fd_client < 0)
		return fd_client;

	*client = ipcconn_new(server->socketname, server->interfacename);
	if(!(*client)) {
		close(fd_client);
		return 1;
	}

	(*client)->lazy_connect = 0;
	(*client)->fd = fd_client;
	return 0;
}

extern int ipcconn_close(ipcconn conn) {
	if(!conn || conn->fd < 0)
		return 1;
	return close(conn->fd);
}

/* message sending/receiving: */

extern int ipcconn_receive_message(ipcconn conn, ipcmessage *msg) {
	char* buffer;
	int length;
	size_t rec_len;

	if(!conn || conn->fd < 0)
		return 1;
#ifdef DEBUG
	printf("IPC: Receiving Message, fd=%d", conn->fd);
#endif

	length = 0;

	/* read one integer (message size) */
	rec_len = recv(conn->fd, (char*) &length, sizeof(int), 0);
#ifdef DEBUG
	printf(", result(size field)=%d, message length shall be %d\n", rec_len, length);
#endif

	if(rec_len < 0)
		return rec_len;

	if(length < sizeof(int))
		return 1;

	/* alloc buffer, will be handed over to ipcmessage */
	buffer = malloc(length);
	if(!buffer)
		return 1;

	/* put length info in the buffer */
	*(int*) buffer = length;
	rec_len = recv(conn->fd, buffer + sizeof(int), length - sizeof(int), 0);

#ifdef DEBUG
	hex_dump(buffer, rec_len + sizeof(int));
#endif

	if(rec_len != length - sizeof(int)) {
		/* we did not receive the needed amount of data... */
		free(buffer);
		return 1;
	}

	*msg = ipcmessage_read(buffer, length);
	free(buffer);
	if(!(*msg)) {
		return 1;
	}

	return 0;
}

extern int ipcconn_send_message(ipcconn conn, ipcmessage msg) {
	size_t snd_len;

	if(!conn || conn->fd < 0 || !msg || !msg->data || *msg->len < 1)
		return 1;

	snd_len = send(conn->fd, msg->data, *msg->len, 0);

#ifdef DEBUG
	printf("IPC: Sent Message, fd=%d, result=%d\n", conn->fd, snd_len);
	hex_dump(msg->data, *msg->len);
#endif

	if(snd_len < 0)
		return snd_len;

	if(snd_len != *msg->len)
		return 1;
	return 0;
}

extern int ipcconn_send_and_receive_message(ipcconn conn, ipcmessage send, ipcmessage *receive) {
	int result = ipcconn_send_message(conn, send);
	if(result)
		return result;
	return ipcconn_receive_message(conn, receive);
}

extern int ipcconn_receive_rpc(ipcconn conn, ipcmessage *receive, int* statusbyte, char** interface, char** procedure) {
	char* socketname = NULL;

	*procedure = NULL;
	*interface = NULL;
	*receive = NULL;

	int err = ipcconn_procedure(conn, NULL, "{bSSSM", statusbyte, procedure, interface, &socketname, receive);
	if(err)
		return err;

	err = strcmp(conn->socketname, socketname);
	if(err) goto onerror;

	/* for now, we do _not_ check interface name. this allows
	   multiple interfaces on the same socket. */

	free(socketname);

	return 0;

onerror:
	if(*procedure) free(*procedure);
	if(*interface) free(*interface);
	if(socketname) free(socketname);
	return err;
}

extern int ipcconn_vparse(ipcmessage msg, const char* signature, va_list args) {
	int err = 0;
	const char *t;
	for(t = signature; *t; t++) {
		if(*t == 'M') {
			/* return full ipcmessage, so do not free it */
			ipcmessage *target = va_arg(args, ipcmessage*);
			*target = msg;
		} else if(*t == 'E') {
			/* standard error checking */
			int iserror;
			int errorcode;
			err = ipcmessage_pop_byte(msg, &iserror);
			if(!err) {
				if(iserror != 0) {
					ipcmessage_pop_string(msg, NULL, NULL);
					ipcmessage_pop_int(msg, &err);
					if(!err)
						err = 1;
				}
			}
		} else if(*t == 'S') {
			/* pop string */
			err = ipcmessage_pop_string(msg, va_arg(args, char**), NULL);
		} else if(*t == 'i') {
			/* pop int */
			err = ipcmessage_pop_int(msg, va_arg(args, int*));
		} else if(*t == 'I') {
			/* peek int */
			err = ipcmessage_peek_int(msg, va_arg(args, int*));
		} else if(*t == '_') {
			/* pop int, do not set any argument */
			err = ipcmessage_pop_int(msg, NULL);
		} else if(*t == 'b') {
			/* pop byte */
			err = ipcmessage_pop_byte(msg, va_arg(args, int*));
		} else if(*t == 'B') {
			/* peek byte */
			err = ipcmessage_peek_byte(msg, va_arg(args, int*));
		} else if(*t == '.') {
			/* pop byte, do not set any argument */
			err = ipcmessage_pop_byte(msg, NULL);
		}
		if(err) break;
	}

	return err;
}

extern int ipcconn_parse(ipcmessage msg, const char* signature, ...) {
	int ret;
	va_list args;
	va_start(args, signature);
	ret = ipcconn_vparse(msg, signature, args);
	va_end(args);
	return ret;
}

extern int ipcconn_vprocedure(ipcconn conn, const char* procedure, const char* signature, va_list args) {
	ipcmessage query = NULL;
	ipcmessage reply = NULL;
	int err = 0;
	int free_reply = 1;
	const char *t;

	if(!conn)
		return 1;

	/* lazy open of client connections */
	if(conn->lazy_connect) {
		err = ipcconn_connect(conn);
		if(err) return err;
		conn->lazy_connect = 0;
	} 

	query = ipcmessage_new();
	if(!query) {
		return 1;
	}

	for(t = signature; *t; t++) {
		if(!reply) {
			if(*t == 'S') {
				/* push C string value */
				err = ipcmessage_push_zstring(query, va_arg(args, const char*));
			} else if(*t == '#') {
				/* push arbitrary buffer, needs two arguments, buffer pointer and buffer size */
				err = ipcmessage_push_string(query, va_arg(args, const char*), va_arg(args, int));
			} else if(*t == 'i') {
				/* push int value */
				err = ipcmessage_push_int(query, va_arg(args, const int));
			} else if(*t == 'b') {
				/* push byte/char value */
				err = ipcmessage_push_byte(query, (const char) va_arg(args, const int));
			} else if(*t == '0') {
				/* push int 0 (TODO: do we need this?) */
				err = ipcmessage_push_int(query, 0);
			} else if(*t == 'P') {
				/* append standard RPC fields */
				err = ipcconn_message_push_rpc(conn, query, procedure, 0);
			} else if(*t == 'p') {
				/* append standard RPC fields, set status byte to 1 */
				err = ipcconn_message_push_rpc(conn, query, procedure, 1);
			} else if(*t == '|') {
				/* issue call, switch to reply mode */
				err = ipcconn_send_and_receive_message(conn, query, &reply);
			} else if(*t == ']') {
				/* issue call, switch to reply mode, don't free reply message */
				err = ipcconn_send_and_receive_message(conn, query, &reply);
				free_reply = 0;
			} else if(*t == '!') {
				/* just issue call */
				err = ipcconn_send_message(conn, query);
			} else if(*t == '{') {
				/* don't send anything, just receive, don't free reply message */
				err = ipcconn_receive_message(conn, &reply);
				free_reply = 0;
			} else if(*t == '<') {
				/* don't send anything, just receive */
				err = ipcconn_receive_message(conn, &reply);
			}
		} else {
			err = ipcconn_vparse(reply, t, args);
			break;
		}
		if(err) break;
	}
	ipcmessage_free(&query);
	if(reply && free_reply)
		ipcmessage_free(&reply);

	va_end(args);

	return err;
}

extern int ipcconn_procedure(ipcconn conn, const char* procedure, const char* signature, ...) {
	int ret;
	va_list args;
	va_start(args, signature);
	ret = ipcconn_vprocedure(conn, procedure, signature, args);
	va_end(args);
	return ret;
}

extern ipcconn ipcconn_get_nameservice() {
	FILE *fp;
	int count, dummy;
	ipcconn ns_conn;
	char *socketname;

	socketname = getenv(NAMESERVICE_ADDRESS_ENVVAR);
	if(!socketname) {
		fp = fopen(NAMESERVICE_ADDRESS_FILE, "r");
		if(!fp) {
			perror("accessing nameservice_address file");
			return NULL;
		}
		count = getline(&socketname, &dummy, fp);
		fclose(fp);
		if(count <= 0) return NULL;
		if(socketname[count - 1] == '\n')
			socketname[count - 1] = '\0';
		ns_conn = NameService_new(socketname);
		free(socketname);
	} else {
		/* we cannot free() a string we got from getenv() */
		ns_conn = NameService_new(socketname);
	}
	return ns_conn;
}
