#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "../vipipc.h"
#include "../vipipc_prototypes.h"

#define SOCKETNAME "/tmp/dummyapplication#0"
#define MAX_CLIENTS 10

#define CONFIGNAME "/tmp/dummyapplication_config.xml"

#define DUMMYMIME "application/dummy"
#define DUMMYURI "file:///dummy"

/* interface definitions: */
IPCPROTO_NEW_SERVER(DummyAppService, "IfaceDummyAppService")
IPCPROTO_NEW_DISPATCHER(DummyAppService, GetAppId)

IPCPROTO_NEW_GENERIC(DummyAppServiceCaller, "IfaceDummyAppService")
IPCPROTO_NEW_CALL(DummyAppServiceCaller, GetAppId, "P|Ei")

ipcconn app, as, ns, uls;
ipcconn client[MAX_CLIENTS];
int application_id;


int app_ping(ipcmessage msg, void* data) {
	/* let's just hope it succeeds: */
	ApplicationService_ReportPing(as, application_id);
	return 0;
}

int dummyapp_getappid(ipcmessage msg, void* data) {
	fprintf(stderr, "got getappid call\n");
	ipcconn_procedure((ipcconn) data, NULL, "ib!", application_id, 0);
	return 0;
}

int run_application(int testmode) {
	int err;
	int ret, i;
	int nfds;
	fd_set rfds;
	struct timeval tv;

	ns = ipcconn_get_nameservice();
	if(!ns) {
		fprintf(stderr, "error creating nameservice ipcconn\n");
		err = 2;
		goto quit;
	}

	as = ApplicationService_new(ns);
	if(!as) {
		fprintf(stderr, "error creating applicationservice ipcconn\n");
		err = 3;
		goto close_ns_and_quit;
	}

	app = Application_new(SOCKETNAME);
	if(!app) {
		fprintf(stderr, "error creating application ipcconn\n");
		err = 4;
		goto close_as_and_quit;
	}
	
	err = ipcconn_bind_and_listen(app, 0777, 2);
	if(err) {
		fprintf(stderr, "error listening+binding to application socket\n");
		err = 5;
		goto close_app_and_quit;
	}

	if(!testmode) {
		err = ApplicationService_ReportStarted(as, application_id, SOCKETNAME);
		if(err) {
			fprintf(stderr, "error notifying application start - we're not known, possibly.\n");
			err = 6;
			goto close_app_and_quit;
		}
	}

	for(i = 0; i < MAX_CLIENTS; i++)
		client[i] = NULL;
	
	while(1) {
		FD_ZERO(&rfds);
		FD_SET(app->fd, &rfds);
		nfds = app->fd;
		for(i = 0; i < MAX_CLIENTS; i++) {
			if(client[i]) {
				FD_SET(client[i]->fd, &rfds);
				if(client[i]->fd > nfds)
					nfds = client[i]->fd;
			}
		}
		nfds++;
		tv.tv_sec = 1200;
		tv.tv_usec = 0;

		ret = select(nfds, &rfds, NULL, NULL, &tv);

		if(ret == -1) {
			perror("select()");
			err = 10;
			break;
		} else if(ret) {
			for(i = 0; i < MAX_CLIENTS; i++) {
				if(client[i] && FD_ISSET(client[i]->fd, &rfds)) {
					/* incoming data on client connectappn */
					char *procedure;
					char *interface;
					int status;
					ipcmessage msg;
					err = ipcconn_receive_rpc(client[i],
							&msg, &status, &interface, &procedure);
					if(!err) {
						fprintf(stderr, "got procedure call: %s:%s\n", interface, procedure);
						if((Application_dispatch_Ping(interface, procedure, msg,
							app_ping, NULL) < 0) &&
						   (DummyAppService_dispatch_GetAppId(interface, procedure, msg,
							dummyapp_getappid, (void *) client[i]) < 0))
						{
							fprintf(stderr, "undispatched procedure call: %s:%s\n",
								interface, procedure);
						}

						if(msg) ipcmessage_free(&msg);
						free(procedure);
						free(interface);
					} else {
						/* close client connectappn on error */
						ipcconn_free(&client[i]);
					}
				}
			}
			if(FD_ISSET(app->fd, &rfds)) {
				/* incoming connection on server socket */

				for(i = 0; i < MAX_CLIENTS; i++)
					if(!client[i]) {
						ipcconn_accept(app, &client[i]);
						break;
					}

				if(i == MAX_CLIENTS) {
					/* no free client slot right now, we accept and
					   drop the connection (probably not clever). */

					ipcconn shortconn;
					ipcconn_accept(app, &shortconn);
					ipcconn_free(&shortconn);
				}
			}
		}
	}

close_app_and_quit:
	ipcconn_free(&app);

close_as_and_quit:
	ipcconn_free(&as);
	
close_ns_and_quit:
	ipcconn_free(&ns);

quit:
	exit(err);
}

void usage(const char* myname) {
	fprintf(stderr, "usage: %s <command> [options...]\n"
			"commands:\n"
			"\tregister <path>\n"
			"\t\tregisters dummyapplication to appman.\n"
			"\t\tthe full path to the dummyapplication binary must\n"
			"\t\tbe given as additional argument.\n"
			"\tstart\n"
			"\t\tissues a dummy LoadUri call which will start\n"
			"\t\tdummyapplication via appman.\n"
			"\tgetappid\n"
			"\t\twill return the application id for a running\n"
			"\t\tinstance of dummyapplication.\n", myname);
	exit(1);
}

int main(int argc, char* argv[]) {
	char *appidstr = getenv("APPLICATION_ID");
	int err, result;

	if(appidstr) {
		/* DEBUG:
		system("/mnt/bin/strace -p $PPID -s 512 -o /mnt/stout 2>/dev/null &");
		sleep(3);
		*/

		application_id = atoi(appidstr); /* cough, cough */
		run_application(0);
	}

	/* we are (most probably) not run by appman */

	if(argc < 2) usage(argv[0]);

	if(!strcmp("testserver", argv[1])) {
		application_id=99;
		run_application(1);
	} else if(!strcmp("register", argv[1])) {
		FILE *fp;

		if(argc < 3) usage(argv[0]);
		fp = fopen(CONFIGNAME, "w");
		if(!fp) {
			perror("open(configfile)");
			exit(1);
		}
		fprintf(fp, "<?xml version=\"1.0\"?>\n"
			"<ApplicationProperties>\n"
			"<PlatformVersion>2.1</PlatformVersion>\n"
			"<Name>dummyapplication</Name>\n"
			"<Run>%s</Run>\n"
			/* "<Icon>/usr/applications/vod/icon_film.png</Icon>\n" */
			"<InitialState>Active</InitialState>\n"
			"<IsPopup>false</IsPopup>\n"
			"<AutoRestart>false</AutoRestart>\n"
			"<StartTimeout>10</StartTimeout>\n"
			"<RegisterUri>mimetype=%s</RegisterUri>\n"
			"<SourceIip>kreatel-app-vod</SourceIip>\n"
			"</ApplicationProperties>\n", argv[2], DUMMYMIME);
		fclose(fp);

		ns = ipcconn_get_nameservice();
		if(!ns) {
			fprintf(stderr, "error creating nameservice ipcconn\n");
			exit(1);
		}

		as = ApplicationService_new(ns);
		if(!as) {
			fprintf(stderr, "error creating applicationservice ipcconn\n");
			ipcconn_free(&ns);
			exit(2);
		}

		err = ApplicationService_RegisterApplication(as, CONFIGNAME);
		if(err) {
			fprintf(stderr, "error when registering application.\n");
		}

		ipcconn_free(&as);
		ipcconn_free(&ns);
		/* unlink(configfile); */
		return err;
	} else if(!strcmp("start", argv[1])) {
		ns = ipcconn_get_nameservice();
		if(!ns) {
			fprintf(stderr, "error creating nameservice ipcconn\n");
			exit(1);
		}

		uls = UriLoaderService_new(ns);
		if(!uls) {
			fprintf(stderr, "error creating uriloaderservice ipcconn\n");
			ipcconn_free(&ns);
			exit(2);
		}

		err = UriLoaderService_LoadUri(uls, DUMMYURI, DUMMYMIME);
		if(err) {
			fprintf(stderr, "error when trying to trigger application start.\n");
			ipcconn_free(&uls);
			exit(3);
		}
		ipcconn_free(&uls);
		ipcconn_free(&ns);
		exit(0);
	} else if(!strcmp("getappid", argv[1])) {
		as = DummyAppServiceCaller_new(SOCKETNAME);
		if(!as) {
			fprintf(stderr, "error creating dummyappservicecaller ipcconn\n");
			exit(1);
		}
		err = DummyAppServiceCaller_GetAppId(as, &result);
		if(err) {
			fprintf(stderr, "IPC call failed.\n");
			exit(2);
		}
		printf("%d\n", result);
		exit(0);
	}
}

