/*
	Copyright (C) 2010 Hans-Werner Hilse <hilse@web.de>
	See file "COPYING" for licence information.
*/
/***************************************************************************
* Example on how to use the vipipcstack library
*
* The resulting program will register itself to the infobroker process
* and listens for changes to the object given as first argument.
*
* For trying it out, run
*
*    example-informationservice config.timezone
*
* To see some action, issue some calls with toish is:
*
*    toish is SetObject config.timezone Europe/Berlin volatile
*    toish is SetObject config.timezone UTC volatile
*    ...
*
* You can see below how a server is managed using select().
*
* The NameService, InformationService and InformationObserver interfaces
* are used.
*
* If you're reading this and you want to learn whats going on "on the
* wire", be sure to also enable debugging output by compiling the
* libvipipcstack with -DDEBUG (see Makefile).
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "../vipipc.h"
#include "../vipipc_prototypes.h"

/* example callback procedure
 *
 * those procedures take the message as first argument and an arbitrary
 * void* as second argument, which you can assign to any data you can
 * point to when dispatching a procedure call. See below on how this is
 * initialized.
 */
int io_onobjectschanged(ipcmessage msg, void* data) {
	int store;
	char *name;
	char *svalue;
	ipcconn is = (ipcconn) data;

	/* for all known answers, parser functions are provided. */
	if(!InformationObserver_parse_OnObjectsChanged(msg, &store, &name)) {
		printf("Got OnObjectsChanged(%d, \"%s\")\n", store, name);
		if(is) {
			if(!InformationService_GetObject(is, name, &svalue)) {
				printf("New value of <%s> is \"%s\"\n", name, svalue);
				free(svalue);
			}
		}
		free(name);
	}
	return 0;
}

int main(int argc, char* argv[]) {
	int err;
	char *svalue;
	int ret;

	/* these will hold the ipcconns */
	ipcconn ns, is, io, iocaller;

	int nfds;
	fd_set rfds;
	struct timeval tv;

	if(argc < 2) {
		printf("error: needs a config variable as first arg (try: \"config.timezone\")\n");
		exit(1);
	}

	/* the following are rather self-explanatory, I think.
	 * just a short note:
	 * on functions returning pointer values, errors are
	 * indicated by returning NULL.
	 * on functions returning int status codes, errors are
	 * indicated by values != 0.
	 */

	/* since the nameservice address is usually determined in a special
	   way (env var NAMESERVICE_ADDRESS or read from file
	   /tmp/nameservice_address, there's a convenient shortcut for this: */
	ns = ipcconn_get_nameservice();
	if(!ns) {
		printf("error creating nameservice ipcconn\n");
		exit(2);
	}

	io = InformationObserver_new("/tmp/settingstest#0");
	if(!io) {
		printf("error creating informationobserver (server) ipcconn\n");
		err = 3;
		goto close_ns_and_quit;
	}
	err = ipcconn_bind_and_listen(io, 0777, 1);
	if(err) {
		printf("error listening+binding to informationobserver socket\n");
		goto close_ns_and_quit;
	}
	
	is = InformationService_new(ns);
	if(!is) {
		printf("error creating informationservice ipcconn\n");
		err = 4;
		goto close_io_and_quit;
	}
	
	err = InformationService_GetObject(is, argv[1], &svalue);
	if(err) {
		printf("error reading <%s>\n", argv[1]);
		goto close_is_and_quit;
	}
	printf("<%s> is set to \"%s\"\n", argv[1], svalue);

	err = InformationService_SubscribeObject(is, argv[1], 1, "/tmp/settingstest#0");
	if(err) {
		printf("error subscribing to <%s> at informationservice\n", argv[1]);
		goto close_is_and_quit;
	}

	/* zero this one out, we'll distinguish connected state on it being != 0 */
	iocaller = 0;

	/* for now, we'll run forever (or 'til you hit CTRL-C) */
	while(1) {
		/* set up fd_set for select(). We only need reading fds. */
		FD_ZERO(&rfds);
		FD_SET(io->fd, &rfds);
		if(iocaller) {
			FD_SET(iocaller->fd, &rfds);
			nfds = (io->fd > iocaller->fd) ? io->fd + 1: iocaller->fd + 1;
		} else {
			nfds = io->fd + 1;
		}

		/* set up timeval for select()s timeout */
		tv.tv_sec = 10;
		tv.tv_usec = 0;

		/* do actual select(), see its man page for all the details */
		ret = select(nfds, &rfds, NULL, NULL, &tv);

		/* we're overly verbose: */
		printf("select(%d, ...) = %d\n", nfds, ret);

		if(ret == -1) {
			/* select returned an error. has to be serious ... */
			perror("select()");
			break;
		} else if(ret) {
			/* file descriptors are ready for reading */
			if(iocaller && FD_ISSET(iocaller->fd, &rfds)) {
				/* incoming data on client connection */
				/* we're listening for IPC-RPC here: */
				char *procedure;
				char *interface;
				int status;
				ipcmessage msg;
				err = ipcconn_receive_rpc(iocaller,
						&msg, &status, &interface, &procedure);
				if(!err) {
					/* all data ready, try to dispatch procedure call: */
					if(InformationObserver_dispatch_OnObjectsChanged(
						interface, procedure, msg,
						io_onobjectschanged, /* callback function! */
						(void*) is           /* data passed through
						                        to callback function */
					     ) < 0)    /* dispatch methods return err < 0
					                  when procedure wasn't for them,
					                  so you can use && to concat
					                  more of them here... */
					{
						/* ... and end here if none of them ran. */
						printf("undispatched procedure call: %s:%s\n",
							interface, procedure);
					}

					if(msg) ipcmessage_free(&msg);
					/* we're in charge of freeing the strings! */
					free(procedure);
					free(interface);
				} else {
					/* close client connection on error */
					ipcconn_free(&iocaller);
				}
			}
			if(FD_ISSET(io->fd, &rfds)) {
				/* ready to read on the server socket: this means that
				   a client is ready and waits to be accept()ed. */

				/* just disconnect existing client session */
				if(iocaller) ipcconn_free(&iocaller);
				
				ipcconn_accept(io, &iocaller);
			}
		}
	}

close_is_and_quit:
	ipcconn_free(&is);

close_io_and_quit:
	ipcconn_free(&io);
	
close_ns_and_quit:
	ipcconn_free(&ns);

	exit(err);
}

