/*
	Copyright (C) 2010 Hans-Werner Hilse <hilse@web.de>
	See file "COPYING" for licence information.
*/
#ifndef __IPCSTACK_IPCCONN
#define __IPCSTACK_IPCCONN

/* we use types defined in those: */
#include <sys/stat.h>
#include <sys/types.h>
#include <stdarg.h>

#define NAMESERVICE_ADDRESS_ENVVAR "NAMESERVICE_ADDRESS"
#define NAMESERVICE_ADDRESS_FILE "/tmp/nameservice_address"

/*******************************************************************************
* IPC MESSAGE BUILDING / PARSING:
*******************************************************************************/

typedef struct {
	char *data;
	int data_bufsize;
	int *len;
	char *ptr;
} ipcmessage_s;

/* this is the main type you'll need: */
typedef ipcmessage_s *ipcmessage;

/* create new message with empty buffer (except initialized int counter) */
extern ipcmessage ipcmessage_new();

/* create new message from existing buffer */
extern ipcmessage ipcmessage_read(char* data, const int length);

/* free message */
extern void ipcmessage_free(ipcmessage* msg);

/* push values onto the message, data is copied for the string types */

extern int ipcmessage_push_int(ipcmessage msg, const int value);
extern int ipcmessage_push_byte(ipcmessage msg, const int value);
extern int ipcmessage_push_string(ipcmessage msg, const char* buffer, const int length);
extern int ipcmessage_push_zstring(ipcmessage msg, const char* buffer);

/* peek values from a message (i.e. do not move pointer),
 * strings will point into message buffer, so do not free()
 * them manually, they are freed when the message is free()d.
 */

extern int ipcmessage_peek_int(ipcmessage msg, int* value);
extern int ipcmessage_peek_byte(ipcmessage msg, int* value);
extern int ipcmessage_peek_string(ipcmessage msg, char** buffer, int* length);

/* pop values from a message (i.e. move pointer) */

extern int ipcmessage_pop_string(ipcmessage msg, char** buffer, int* length);
extern int ipcmessage_pop_byte(ipcmessage msg, int* value);
extern int ipcmessage_pop_int(ipcmessage msg, int* value);



/*******************************************************************************
* IPC CONNECTION HANDLING:
*******************************************************************************/

typedef struct {
	int fd;
	int lazy_connect;
	char* socketname;
	char* interfacename;
} ipcconn_s;

/* this is the main type you'll need: */
typedef ipcconn_s *ipcconn;

/* create a new connection "object" */
extern ipcconn ipcconn_new(const char* socketname, const char* interfacename);

/* free resources & close open socket for connection */
extern void ipcconn_free(ipcconn* conn);

/* simple helper to push connection info and an RPC call onto a ipcmessage stack */
extern int ipcconn_message_push_rpc(ipcconn conn, ipcmessage msg, const char* procedure, const char statusbyte);

/* socket management: */
extern int ipcconn_connect(ipcconn client);
extern int ipcconn_bind(ipcconn server, mode_t socketpermissions);
extern int ipcconn_listen(ipcconn server, int backlog);
extern int ipcconn_accept(ipcconn server, ipcconn *client);
extern int ipcconn_close(ipcconn conn);
/* convenience combo call: */
extern int ipcconn_bind_and_listen(ipcconn server, mode_t socketpermissions, int backlog);



/* this falls a bit out of the category "connection", but it's in this namespace nevertheless: */


/* message sending/receiving: */
extern int ipcconn_receive_message(ipcconn conn, ipcmessage *msg);
extern int ipcconn_send_message(ipcconn conn, ipcmessage msg);

/* convenience combo call */
extern int ipcconn_send_and_receive_message(ipcconn, ipcmessage send, ipcmessage *receive);

/* parser for an RPC message footer (i.e. stack header) */
extern int ipcconn_receive_rpc(ipcconn conn, ipcmessage *receive, int* statusbyte, char** interface, char** procedure);

/* universal parser for IPC messages, reading values according to a signature string */
extern int ipcconn_vparse(ipcmessage msg, const char* signature, va_list args);
extern int ipcconn_parse(ipcmessage msg, const char* signature, ...);

/* message construction, passing, receiving and reply parsing functions, adhering to a signature string */
extern int ipcconn_vprocedure(ipcconn conn, const char* procedure, const char* signature, va_list args);
extern int ipcconn_procedure(ipcconn conn, const char* procedure, const char* signature, ...);

/* helper to get nameservice ipcconn */
extern ipcconn ipcconn_get_nameservice();

/*******************************************************************************
* MACROS FOR IPC PROTOTYPES:  (yes, it's a bit ugly, but makes protyping fast)
*******************************************************************************/

#define IPCPROTO_NEW_GENERIC_D(protoname) \
	extern ipcconn protoname##_new(const char* socketname);
#define IPCPROTO_NEW_GENERIC(protoname, interfacename) \
	extern ipcconn protoname##_new(const char* socketname) { \
		return ipcconn_new(socketname, interfacename); \
	}

#define IPCPROTO_NEW_SERVICE_D(protoname) \
	extern ipcconn protoname##_new(ipcconn nameserverconn);
#define IPCPROTO_NEW_SERVICE(protoname, servicename, interfacename) \
	extern ipcconn protoname##_new(ipcconn nameserverconn) { \
		ipcconn result; \
		char *socketname; \
		if(NameService_LookupObject(nameserverconn, servicename, &socketname)) \
			return 0; \
		if(!socketname) \
			return 0; \
		result = ipcconn_new(socketname, interfacename); \
		free(socketname); \
		return result; \
	}

#define IPCPROTO_NEW_CALL_D(protoname, callname) \
	extern int protoname##_##callname(ipcconn myconn, ...);

#define IPCPROTO_NEW_CALL(protoname, callname, signature) \
	extern int protoname##_##callname(ipcconn myconn, ...) { \
		int ret; \
		va_list argp; \
		va_start(argp, myconn); \
		ret = ipcconn_vprocedure(myconn, #callname, signature, argp); \
		va_end(argp); \
		return ret; \
	}

#define IPCPROTO_NEW_SERVER_D(protoname) \
	extern ipcconn protoname##_new(const char* socketname); \
	extern int protoname##_check_interface(const char* interface);
#define IPCPROTO_NEW_SERVER(protoname, interfacename) \
	extern ipcconn protoname##_new(const char* socketname) { \
		return ipcconn_new(socketname, interfacename); \
	} \
	extern int protoname##_check_interface(const char* interface) { \
		return strcmp(interfacename, interface); \
	}

#define IPCPROTO_NEW_PARSER_D(protoname, callname) \
	extern int protoname##_parse_##callname(ipcmessage msg, ...);
#define IPCPROTO_NEW_PARSER(protoname, callname, signature) \
	extern int protoname##_parse_##callname(ipcmessage msg, ...) { \
		int ret; \
		va_list argp; \
		va_start(argp, msg); \
		ret = ipcconn_vparse(msg, signature, argp); \
		va_end(argp); \
		return ret; \
	}
#define IPCPROTO_NEW_DISPATCHER_D(protoname, callname) \
	extern int protoname##_dispatch_##callname(\
		const char* interface, \
		const char* procedure, \
		ipcmessage msg, \
		int (*handler) (ipcmessage msg, void* data), \
		void* userdata);
#define IPCPROTO_NEW_DISPATCHER(protoname, callname) \
	extern int protoname##_dispatch_##callname(\
		const char* interface, \
		const char* procedure, \
		ipcmessage msg, \
		int (*handler) (ipcmessage msg, void* data), \
		void* userdata) \
	{ \
		if(protoname##_check_interface(interface)) return -1; \
		if(strcmp(#callname, procedure)) return -1; \
		return handler(msg, userdata); \
	}

#endif /* __IPCSTACK_IPCCONN */
